#include "../../include/CD/cd_internal.hpp"
#include <PSX/System/IOPorts/interrupt_io.hpp>
#include <PSX/System/syscalls.h>

#include <stdio.h>

namespace JabyEngine {
    namespace CD {
        namespace internal {
            struct Mode : public ComplexBitMap<uint8_t> {
                static constexpr auto DoubleSpeed    = Bit<uint8_t>(7);
                static constexpr auto SingleSpeed    = !DoubleSpeed;
                static constexpr auto XADPCM         = Bit<uint8_t>(6);
                static constexpr auto WholeSector    = Bit<uint8_t>(5);
                static constexpr auto DataSector     = !WholeSector;
                static constexpr auto UseXAFilter    = Bit<uint8_t>(3);
                static constexpr auto AudioPlayIRQ   = Bit<uint8_t>(2);
                static constexpr auto AutoPauseTrack = Bit<uint8_t>(1);
                static constexpr auto CDDA           = Bit<uint8_t>(0);
            };

            static constexpr auto DataSectorMode = Mode::with(Mode::DoubleSpeed, Mode::DataSector);

            static InterruptVerifierResult interrupt_verifier() {
                if(Interrupt::is_irq(Interrupt::CDROM)) {
                    const uint8_t old_idx = (CD_IO::IndexStatus.read() & 0x3);

                    CD_IO::PortIndex1::change_to();

                    last_interrupt.write(CD_IO::Interrupt::get_type(CD_IO::PortIndex1::InterruptFlagRegister));
                    CD_IO::Interrupt::ack(CD_IO::PortIndex1::InterruptFlagRegister);

                    CD_IO::IndexStatus.write({old_idx});
                    return InterruptVerifierResult::ExecuteHandler;
                }
                
                else {
                    return InterruptVerifierResult::SkipHandler;
                }
            }

            static void interrupt_handler(uint32_t) {
                Interrupt::ack_irq(Interrupt::CDROM);
                __syscall_ReturnFromException();
            }

            VolatilePOD<CD_IO::Interrupt::Type> last_interrupt{CD_IO::Interrupt::Type::None};
            InterrupCallback callback = {
                .next              = nullptr,
                .handler_function  = reinterpret_cast<InterruptHandler>(interrupt_handler),
                .verifier_function = interrupt_verifier
            };

            State read_file(FileInfo file_info, const SectorBufferAllocator& buffer_allocator) {
                CD_IO::PortIndex0::change_to();
                Command::send_wait(CD_IO::PortIndex0::CommandFifo, CD_IO::PortIndex0::ParameterFifo, CD_IO::Command::SetMode, DataSectorMode);

                printf("I'm not fully implemented! %s\n", __FUNCTION__);
                return State::Error;
            }
        }
    }
}